
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "vulkan-enumtypes.h"

#include <gst/vulkan/vulkan.h>

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gstvkbarrier.h" */
GType
gst_vulkan_barrier_type_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_VULKAN_BARRIER_NONE), "GST_VULKAN_BARRIER_NONE", "none" },
      { C_ENUM (GST_VULKAN_BARRIER_TYPE_MEMORY), "GST_VULKAN_BARRIER_TYPE_MEMORY", "type-memory" },
      { C_ENUM (GST_VULKAN_BARRIER_TYPE_BUFFER), "GST_VULKAN_BARRIER_TYPE_BUFFER", "type-buffer" },
      { C_ENUM (GST_VULKAN_BARRIER_TYPE_IMAGE), "GST_VULKAN_BARRIER_TYPE_IMAGE", "type-image" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstVulkanBarrierType", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}
GType
gst_vulkan_barrier_flags_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_VULKAN_BARRIER_FLAG_NONE), "GST_VULKAN_BARRIER_FLAG_NONE", "none" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstVulkanBarrierFlags", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstvkdisplay.h" */
GType
gst_vulkan_display_type_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GFlagsValue values[] = {
      { C_FLAGS (GST_VULKAN_DISPLAY_TYPE_NONE), "GST_VULKAN_DISPLAY_TYPE_NONE", "none" },
      { C_FLAGS (GST_VULKAN_DISPLAY_TYPE_XCB), "GST_VULKAN_DISPLAY_TYPE_XCB", "xcb" },
      { C_FLAGS (GST_VULKAN_DISPLAY_TYPE_WAYLAND), "GST_VULKAN_DISPLAY_TYPE_WAYLAND", "wayland" },
      { C_FLAGS (GST_VULKAN_DISPLAY_TYPE_COCOA), "GST_VULKAN_DISPLAY_TYPE_COCOA", "cocoa" },
      { C_FLAGS (GST_VULKAN_DISPLAY_TYPE_IOS), "GST_VULKAN_DISPLAY_TYPE_IOS", "ios" },
      { C_FLAGS (GST_VULKAN_DISPLAY_TYPE_WIN32), "GST_VULKAN_DISPLAY_TYPE_WIN32", "win32" },
      { C_FLAGS (GST_VULKAN_DISPLAY_TYPE_ANDROID), "GST_VULKAN_DISPLAY_TYPE_ANDROID", "android" },
      { C_FLAGS (GST_VULKAN_DISPLAY_TYPE_ANY), "GST_VULKAN_DISPLAY_TYPE_ANY", "any" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_flags_register_static ("GstVulkanDisplayType", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstvkerror.h" */
GType
gst_vulkan_error_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_VULKAN_FAILED), "GST_VULKAN_FAILED", "failed" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstVulkanError", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstvkformat.h" */
GType
gst_vulkan_format_scaling_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_VULKAN_FORMAT_SCALING_UNORM), "GST_VULKAN_FORMAT_SCALING_UNORM", "unorm" },
      { C_ENUM (GST_VULKAN_FORMAT_SCALING_SNORM), "GST_VULKAN_FORMAT_SCALING_SNORM", "snorm" },
      { C_ENUM (GST_VULKAN_FORMAT_SCALING_USCALED), "GST_VULKAN_FORMAT_SCALING_USCALED", "uscaled" },
      { C_ENUM (GST_VULKAN_FORMAT_SCALING_SSCALED), "GST_VULKAN_FORMAT_SCALING_SSCALED", "sscaled" },
      { C_ENUM (GST_VULKAN_FORMAT_SCALING_UINT), "GST_VULKAN_FORMAT_SCALING_UINT", "uint" },
      { C_ENUM (GST_VULKAN_FORMAT_SCALING_SINT), "GST_VULKAN_FORMAT_SCALING_SINT", "sint" },
      { C_ENUM (GST_VULKAN_FORMAT_SCALING_SRGB), "GST_VULKAN_FORMAT_SCALING_SRGB", "srgb" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstVulkanFormatScaling", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}
GType
gst_vulkan_format_flags_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GFlagsValue values[] = {
      { C_FLAGS (GST_VULKAN_FORMAT_FLAG_YUV), "GST_VULKAN_FORMAT_FLAG_YUV", "yuv" },
      { C_FLAGS (GST_VULKAN_FORMAT_FLAG_RGB), "GST_VULKAN_FORMAT_FLAG_RGB", "rgb" },
      { C_FLAGS (GST_VULKAN_FORMAT_FLAG_ALPHA), "GST_VULKAN_FORMAT_FLAG_ALPHA", "alpha" },
      { C_FLAGS (GST_VULKAN_FORMAT_FLAG_LE), "GST_VULKAN_FORMAT_FLAG_LE", "le" },
      { C_FLAGS (GST_VULKAN_FORMAT_FLAG_COMPLEX), "GST_VULKAN_FORMAT_FLAG_COMPLEX", "complex" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_flags_register_static ("GstVulkanFormatFlags", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstvkhandle.h" */
GType
gst_vulkan_handle_type_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_VULKAN_HANDLE_TYPE_DESCRIPTOR_SET_LAYOUT), "GST_VULKAN_HANDLE_TYPE_DESCRIPTOR_SET_LAYOUT", "descriptor-set-layout" },
      { C_ENUM (GST_VULKAN_HANDLE_TYPE_PIPELINE_LAYOUT), "GST_VULKAN_HANDLE_TYPE_PIPELINE_LAYOUT", "pipeline-layout" },
      { C_ENUM (GST_VULKAN_HANDLE_TYPE_PIPELINE), "GST_VULKAN_HANDLE_TYPE_PIPELINE", "pipeline" },
      { C_ENUM (GST_VULKAN_HANDLE_TYPE_RENDER_PASS), "GST_VULKAN_HANDLE_TYPE_RENDER_PASS", "render-pass" },
      { C_ENUM (GST_VULKAN_HANDLE_TYPE_SAMPLER), "GST_VULKAN_HANDLE_TYPE_SAMPLER", "sampler" },
      { C_ENUM (GST_VULKAN_HANDLE_TYPE_FRAMEBUFFER), "GST_VULKAN_HANDLE_TYPE_FRAMEBUFFER", "framebuffer" },
      { C_ENUM (GST_VULKAN_HANDLE_TYPE_SHADER), "GST_VULKAN_HANDLE_TYPE_SHADER", "shader" },
      { C_ENUM (GST_VULKAN_HANDLE_TYPE_VIDEO_SESSION), "GST_VULKAN_HANDLE_TYPE_VIDEO_SESSION", "video-session" },
      { C_ENUM (GST_VULKAN_HANDLE_TYPE_VIDEO_SESSION_PARAMETERS), "GST_VULKAN_HANDLE_TYPE_VIDEO_SESSION_PARAMETERS", "video-session-parameters" },
      { C_ENUM (GST_VULKAN_HANDLE_TYPE_SAMPLER_YCBCR_CONVERSION), "GST_VULKAN_HANDLE_TYPE_SAMPLER_YCBCR_CONVERSION", "sampler-ycbcr-conversion" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstVulkanHandleType", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstvkwindow.h" */
GType
gst_vulkan_window_error_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_VULKAN_WINDOW_ERROR_FAILED), "GST_VULKAN_WINDOW_ERROR_FAILED", "failed" },
      { C_ENUM (GST_VULKAN_WINDOW_ERROR_OLD_LIBS), "GST_VULKAN_WINDOW_ERROR_OLD_LIBS", "old-libs" },
      { C_ENUM (GST_VULKAN_WINDOW_ERROR_RESOURCE_UNAVAILABLE), "GST_VULKAN_WINDOW_ERROR_RESOURCE_UNAVAILABLE", "resource-unavailable" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstVulkanWindowError", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* Generated data ends here */

