
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "gstsrt-enumtypes.h"
#include "gstsrt-enums.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gstsrt-enums.h" */

GType
gst_srt_connection_mode_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GST_SRT_CONNECTION_MODE_NONE), "GST_SRT_CONNECTION_MODE_NONE", "none" },
        { C_ENUM (GST_SRT_CONNECTION_MODE_CALLER), "GST_SRT_CONNECTION_MODE_CALLER", "caller" },
        { C_ENUM (GST_SRT_CONNECTION_MODE_LISTENER), "GST_SRT_CONNECTION_MODE_LISTENER", "listener" },
        { C_ENUM (GST_SRT_CONNECTION_MODE_RENDEZVOUS), "GST_SRT_CONNECTION_MODE_RENDEZVOUS", "rendezvous" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GstSRTConnectionMode"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gst_srt_key_length_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GST_SRT_KEY_LENGTH_NO_KEY), "GST_SRT_KEY_LENGTH_NO_KEY", "no-key" },
        { C_ENUM (GST_SRT_KEY_LENGTH_0), "GST_SRT_KEY_LENGTH_0", "0" },
        { C_ENUM (GST_SRT_KEY_LENGTH_16), "GST_SRT_KEY_LENGTH_16", "16" },
        { C_ENUM (GST_SRT_KEY_LENGTH_24), "GST_SRT_KEY_LENGTH_24", "24" },
        { C_ENUM (GST_SRT_KEY_LENGTH_32), "GST_SRT_KEY_LENGTH_32", "32" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GstSRTKeyLength"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

