
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "tag-enumtypes.h"
#include "gsttagdemux.h"
#include "tag.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gsttagdemux.h" */

GType
gst_tag_demux_result_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GST_TAG_DEMUX_RESULT_BROKEN_TAG), "GST_TAG_DEMUX_RESULT_BROKEN_TAG", "broken-tag" },
        { C_ENUM (GST_TAG_DEMUX_RESULT_AGAIN), "GST_TAG_DEMUX_RESULT_AGAIN", "again" },
        { C_ENUM (GST_TAG_DEMUX_RESULT_OK), "GST_TAG_DEMUX_RESULT_OK", "ok" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GstTagDemuxResult"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "tag.h" */

GType
gst_tag_image_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GST_TAG_IMAGE_TYPE_NONE), "GST_TAG_IMAGE_TYPE_NONE", "none" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_UNDEFINED), "GST_TAG_IMAGE_TYPE_UNDEFINED", "undefined" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_FRONT_COVER), "GST_TAG_IMAGE_TYPE_FRONT_COVER", "front-cover" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_BACK_COVER), "GST_TAG_IMAGE_TYPE_BACK_COVER", "back-cover" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_LEAFLET_PAGE), "GST_TAG_IMAGE_TYPE_LEAFLET_PAGE", "leaflet-page" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_MEDIUM), "GST_TAG_IMAGE_TYPE_MEDIUM", "medium" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_LEAD_ARTIST), "GST_TAG_IMAGE_TYPE_LEAD_ARTIST", "lead-artist" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_ARTIST), "GST_TAG_IMAGE_TYPE_ARTIST", "artist" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_CONDUCTOR), "GST_TAG_IMAGE_TYPE_CONDUCTOR", "conductor" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_BAND_ORCHESTRA), "GST_TAG_IMAGE_TYPE_BAND_ORCHESTRA", "band-orchestra" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_COMPOSER), "GST_TAG_IMAGE_TYPE_COMPOSER", "composer" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_LYRICIST), "GST_TAG_IMAGE_TYPE_LYRICIST", "lyricist" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_RECORDING_LOCATION), "GST_TAG_IMAGE_TYPE_RECORDING_LOCATION", "recording-location" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_DURING_RECORDING), "GST_TAG_IMAGE_TYPE_DURING_RECORDING", "during-recording" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_DURING_PERFORMANCE), "GST_TAG_IMAGE_TYPE_DURING_PERFORMANCE", "during-performance" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_VIDEO_CAPTURE), "GST_TAG_IMAGE_TYPE_VIDEO_CAPTURE", "video-capture" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_FISH), "GST_TAG_IMAGE_TYPE_FISH", "fish" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_ILLUSTRATION), "GST_TAG_IMAGE_TYPE_ILLUSTRATION", "illustration" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_BAND_ARTIST_LOGO), "GST_TAG_IMAGE_TYPE_BAND_ARTIST_LOGO", "band-artist-logo" },
        { C_ENUM (GST_TAG_IMAGE_TYPE_PUBLISHER_STUDIO_LOGO), "GST_TAG_IMAGE_TYPE_PUBLISHER_STUDIO_LOGO", "publisher-studio-logo" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GstTagImageType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gst_tag_license_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GST_TAG_LICENSE_PERMITS_REPRODUCTION), "GST_TAG_LICENSE_PERMITS_REPRODUCTION", "permits-reproduction" },
        { C_FLAGS (GST_TAG_LICENSE_PERMITS_DISTRIBUTION), "GST_TAG_LICENSE_PERMITS_DISTRIBUTION", "permits-distribution" },
        { C_FLAGS (GST_TAG_LICENSE_PERMITS_DERIVATIVE_WORKS), "GST_TAG_LICENSE_PERMITS_DERIVATIVE_WORKS", "permits-derivative-works" },
        { C_FLAGS (GST_TAG_LICENSE_PERMITS_SHARING), "GST_TAG_LICENSE_PERMITS_SHARING", "permits-sharing" },
        { C_FLAGS (GST_TAG_LICENSE_REQUIRES_NOTICE), "GST_TAG_LICENSE_REQUIRES_NOTICE", "requires-notice" },
        { C_FLAGS (GST_TAG_LICENSE_REQUIRES_ATTRIBUTION), "GST_TAG_LICENSE_REQUIRES_ATTRIBUTION", "requires-attribution" },
        { C_FLAGS (GST_TAG_LICENSE_REQUIRES_SHARE_ALIKE), "GST_TAG_LICENSE_REQUIRES_SHARE_ALIKE", "requires-share-alike" },
        { C_FLAGS (GST_TAG_LICENSE_REQUIRES_SOURCE_CODE), "GST_TAG_LICENSE_REQUIRES_SOURCE_CODE", "requires-source-code" },
        { C_FLAGS (GST_TAG_LICENSE_REQUIRES_COPYLEFT), "GST_TAG_LICENSE_REQUIRES_COPYLEFT", "requires-copyleft" },
        { C_FLAGS (GST_TAG_LICENSE_REQUIRES_LESSER_COPYLEFT), "GST_TAG_LICENSE_REQUIRES_LESSER_COPYLEFT", "requires-lesser-copyleft" },
        { C_FLAGS (GST_TAG_LICENSE_PROHIBITS_COMMERCIAL_USE), "GST_TAG_LICENSE_PROHIBITS_COMMERCIAL_USE", "prohibits-commercial-use" },
        { C_FLAGS (GST_TAG_LICENSE_PROHIBITS_HIGH_INCOME_NATION_USE), "GST_TAG_LICENSE_PROHIBITS_HIGH_INCOME_NATION_USE", "prohibits-high-income-nation-use" },
        { C_FLAGS (GST_TAG_LICENSE_CREATIVE_COMMONS_LICENSE), "GST_TAG_LICENSE_CREATIVE_COMMONS_LICENSE", "creative-commons-license" },
        { C_FLAGS (GST_TAG_LICENSE_FREE_SOFTWARE_FOUNDATION_LICENSE), "GST_TAG_LICENSE_FREE_SOFTWARE_FOUNDATION_LICENSE", "free-software-foundation-license" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GstTagLicenseFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

