#[allow(clippy::wildcard_imports)]
use super::*;
#[allow(clippy::unreadable_literal)]
pub(crate) fn generate(r: &mut HashMap<&'static str, Resource>) {
r.insert("css/graphviz.svg.css",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/css/graphviz.svg.css"),1781805918,"text/css"));
r.insert("dist/bundle.css",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/dist/bundle.css"),1781805918,"text/css"));
r.insert("dist/bundle.js",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/dist/bundle.js"),1781805918,"text/javascript"));
r.insert("dist/bundle.js.LICENSE.txt",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/dist/bundle.js.LICENSE.txt"),1781805918,"text/plain"));
r.insert("images/favicon-120.png",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/favicon-120.png"),1781805918,"image/png"));
r.insert("images/favicon-128.png",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/favicon-128.png"),1781805918,"image/png"));
r.insert("images/favicon-152.png",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/favicon-152.png"),1781805918,"image/png"));
r.insert("images/favicon-16.png",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/favicon-16.png"),1781805918,"image/png"));
r.insert("images/favicon-180.png",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/favicon-180.png"),1781805918,"image/png"));
r.insert("images/favicon-192.png",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/favicon-192.png"),1781805918,"image/png"));
r.insert("images/favicon-196.png",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/favicon-196.png"),1781805918,"image/png"));
r.insert("images/favicon-228.png",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/favicon-228.png"),1781805918,"image/png"));
r.insert("images/favicon-32.png",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/favicon-32.png"),1781805918,"image/png"));
r.insert("images/favicon-57.png",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/favicon-57.png"),1781805918,"image/png"));
r.insert("images/favicon-76.png",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/favicon-76.png"),1781805918,"image/png"));
r.insert("images/favicon-96.png",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/favicon-96.png"),1781805918,"image/png"));
r.insert("images/favicon.png",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/favicon.png"),1781805918,"image/png"));
r.insert("images/gst-dots-viewer-video.jpeg",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/gst-dots-viewer-video.jpeg"),1781805918,"image/jpeg"));
r.insert("images/gst-dots-viewer.jpeg",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/images/gst-dots-viewer.jpeg"),1781805918,"image/jpeg"));
r.insert("index.html",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/index.html"),1781805918,"text/html"));
r.insert("js/gstdots.js",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/js/gstdots.js"),1781805918,"text/javascript"));
r.insert("js/jquery.graphviz.svg.js",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/js/jquery.graphviz.svg.js"),1781805918,"text/javascript"));
r.insert("js/pipeline-navigation.js",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/js/pipeline-navigation.js"),1781805918,"text/javascript"));
r.insert("js/svg-overlay-manager.js",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/js/svg-overlay-manager.js"),1781805918,"text/javascript"));
r.insert("js/text-ellipsizer.js",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/js/text-ellipsizer.js"),1781805918,"text/javascript"));
r.insert("js/tooltip.js",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/js/tooltip.js"),1781805918,"text/javascript"));
r.insert("overlay.html",n(i!("/build/gstreamer/src/gstreamer/subprojects/gst-devtools/dots-viewer/static/overlay.html"),1781805918,"text/html"));
}
