
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "gstsrtp-enumtypes.h"
#include "gstsrtpenums.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gstsrtpenums.h" */

GType
gst_srtp_cipher_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GST_SRTP_CIPHER_NULL), "GST_SRTP_CIPHER_NULL", "null" },
        { C_ENUM (GST_SRTP_CIPHER_AES_128_ICM), "GST_SRTP_CIPHER_AES_128_ICM", "aes-128-icm" },
        { C_ENUM (GST_SRTP_CIPHER_AES_256_ICM), "GST_SRTP_CIPHER_AES_256_ICM", "aes-256-icm" },
        { C_ENUM (GST_SRTP_CIPHER_AES_128_GCM), "GST_SRTP_CIPHER_AES_128_GCM", "aes-128-gcm" },
        { C_ENUM (GST_SRTP_CIPHER_AES_256_GCM), "GST_SRTP_CIPHER_AES_256_GCM", "aes-256-gcm" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GstSrtpCipherType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gst_srtp_auth_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GST_SRTP_AUTH_NULL), "GST_SRTP_AUTH_NULL", "null" },
        { C_ENUM (GST_SRTP_AUTH_HMAC_SHA1_32), "GST_SRTP_AUTH_HMAC_SHA1_32", "hmac-sha1-32" },
        { C_ENUM (GST_SRTP_AUTH_HMAC_SHA1_80), "GST_SRTP_AUTH_HMAC_SHA1_80", "hmac-sha1-80" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GstSrtpAuthType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

