
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */


#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "gst-validate-enum-types.h"
#include <gst/validate/validate.h>
#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)
/* enumerations from "gst-validate-enums.h" */
GType
gst_validate_reporting_details_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(GST_VALIDATE_SHOW_UNKNOWN), "GST_VALIDATE_SHOW_UNKNOWN", "unknown" },
    { C_ENUM(GST_VALIDATE_SHOW_NONE), "GST_VALIDATE_SHOW_NONE", "none" },
    { C_ENUM(GST_VALIDATE_SHOW_SYNTHETIC), "GST_VALIDATE_SHOW_SYNTHETIC", "synthetic" },
    { C_ENUM(GST_VALIDATE_SHOW_SUBCHAIN), "GST_VALIDATE_SHOW_SUBCHAIN", "subchain" },
    { C_ENUM(GST_VALIDATE_SHOW_MONITOR), "GST_VALIDATE_SHOW_MONITOR", "monitor" },
    { C_ENUM(GST_VALIDATE_SHOW_ALL), "GST_VALIDATE_SHOW_ALL", "all" },
    { C_ENUM(GST_VALIDATE_SHOW_SMART), "GST_VALIDATE_SHOW_SMART", "smart" },
    { C_ENUM(GST_VALIDATE_SHOW_COUNT), "GST_VALIDATE_SHOW_COUNT", "count" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("GstValidateReportingDetails", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
gst_validate_verbosity_flags_get_type (void)
{
  static gsize id = 0;
  static const GFlagsValue values[] = {
    { C_FLAGS(GST_VALIDATE_VERBOSITY_NONE), "GST_VALIDATE_VERBOSITY_NONE", "none" },
    { C_FLAGS(GST_VALIDATE_VERBOSITY_POSITION), "GST_VALIDATE_VERBOSITY_POSITION", "position" },
    { C_FLAGS(GST_VALIDATE_VERBOSITY_MESSAGES), "GST_VALIDATE_VERBOSITY_MESSAGES", "messages" },
    { C_FLAGS(GST_VALIDATE_VERBOSITY_PROPS_CHANGES), "GST_VALIDATE_VERBOSITY_PROPS_CHANGES", "props-changes" },
    { C_FLAGS(GST_VALIDATE_VERBOSITY_NEW_ELEMENTS), "GST_VALIDATE_VERBOSITY_NEW_ELEMENTS", "new-elements" },
    { C_FLAGS(GST_VALIDATE_VERBOSITY_ALL), "GST_VALIDATE_VERBOSITY_ALL", "all" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_flags_register_static ("GstValidateVerbosityFlags", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
gst_validate_structure_resolve_variables_flags_get_type (void)
{
  static gsize id = 0;
  static const GFlagsValue values[] = {
    { C_FLAGS(GST_VALIDATE_STRUCTURE_RESOLVE_VARIABLES_ALL), "GST_VALIDATE_STRUCTURE_RESOLVE_VARIABLES_ALL", "all" },
    { C_FLAGS(GST_VALIDATE_STRUCTURE_RESOLVE_VARIABLES_LOCAL_ONLY), "GST_VALIDATE_STRUCTURE_RESOLVE_VARIABLES_LOCAL_ONLY", "local-only" },
    { C_FLAGS(GST_VALIDATE_STRUCTURE_RESOLVE_VARIABLES_NO_FAILURE), "GST_VALIDATE_STRUCTURE_RESOLVE_VARIABLES_NO_FAILURE", "no-failure" },
    { C_FLAGS(GST_VALIDATE_STRUCTURE_RESOLVE_VARIABLES_NO_EXPRESSION), "GST_VALIDATE_STRUCTURE_RESOLVE_VARIABLES_NO_EXPRESSION", "no-expression" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_flags_register_static ("GstValidateStructureResolveVariablesFlags", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
gst_validate_object_set_property_flags_get_type (void)
{
  static gsize id = 0;
  static const GFlagsValue values[] = {
    { C_FLAGS(GST_VALIDATE_OBJECT_SET_PROPERTY_FLAGS_OPTIONAL), "GST_VALIDATE_OBJECT_SET_PROPERTY_FLAGS_OPTIONAL", "optional" },
    { C_FLAGS(GST_VALIDATE_OBJECT_SET_PROPERTY_FLAGS_NO_VALUE_CHECK), "GST_VALIDATE_OBJECT_SET_PROPERTY_FLAGS_NO_VALUE_CHECK", "no-value-check" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_flags_register_static ("GstValidateObjectSetPropertyFlags", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "gst-validate-report.h" */
GType
gst_validate_debug_flags_get_type (void)
{
  static gsize id = 0;
  static const GFlagsValue values[] = {
    { C_FLAGS(GST_VALIDATE_FATAL_DEFAULT), "GST_VALIDATE_FATAL_DEFAULT", "fatal-default" },
    { C_FLAGS(GST_VALIDATE_FATAL_ISSUES), "GST_VALIDATE_FATAL_ISSUES", "fatal-issues" },
    { C_FLAGS(GST_VALIDATE_FATAL_WARNINGS), "GST_VALIDATE_FATAL_WARNINGS", "fatal-warnings" },
    { C_FLAGS(GST_VALIDATE_FATAL_CRITICALS), "GST_VALIDATE_FATAL_CRITICALS", "fatal-criticals" },
    { C_FLAGS(GST_VALIDATE_PRINT_ISSUES), "GST_VALIDATE_PRINT_ISSUES", "print-issues" },
    { C_FLAGS(GST_VALIDATE_PRINT_WARNINGS), "GST_VALIDATE_PRINT_WARNINGS", "print-warnings" },
    { C_FLAGS(GST_VALIDATE_PRINT_CRITICALS), "GST_VALIDATE_PRINT_CRITICALS", "print-criticals" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_flags_register_static ("GstValidateDebugFlags", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
gst_validate_report_level_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(GST_VALIDATE_REPORT_LEVEL_CRITICAL), "GST_VALIDATE_REPORT_LEVEL_CRITICAL", "critical" },
    { C_ENUM(GST_VALIDATE_REPORT_LEVEL_WARNING), "GST_VALIDATE_REPORT_LEVEL_WARNING", "warning" },
    { C_ENUM(GST_VALIDATE_REPORT_LEVEL_ISSUE), "GST_VALIDATE_REPORT_LEVEL_ISSUE", "issue" },
    { C_ENUM(GST_VALIDATE_REPORT_LEVEL_IGNORE), "GST_VALIDATE_REPORT_LEVEL_IGNORE", "ignore" },
    { C_ENUM(GST_VALIDATE_REPORT_LEVEL_UNKNOWN), "GST_VALIDATE_REPORT_LEVEL_UNKNOWN", "unknown" },
    { C_ENUM(GST_VALIDATE_REPORT_LEVEL_EXPECTED), "GST_VALIDATE_REPORT_LEVEL_EXPECTED", "expected" },
    { C_ENUM(GST_VALIDATE_REPORT_LEVEL_NUM_ENTRIES), "GST_VALIDATE_REPORT_LEVEL_NUM_ENTRIES", "num-entries" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("GstValidateReportLevel", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
gst_validate_issue_flags_get_type (void)
{
  static gsize id = 0;
  static const GFlagsValue values[] = {
    { C_FLAGS(GST_VALIDATE_ISSUE_FLAGS_NONE), "GST_VALIDATE_ISSUE_FLAGS_NONE", "none" },
    { C_FLAGS(GST_VALIDATE_ISSUE_FLAGS_FULL_DETAILS), "GST_VALIDATE_ISSUE_FLAGS_FULL_DETAILS", "full-details" },
    { C_FLAGS(GST_VALIDATE_ISSUE_FLAGS_NO_BACKTRACE), "GST_VALIDATE_ISSUE_FLAGS_NO_BACKTRACE", "no-backtrace" },
    { C_FLAGS(GST_VALIDATE_ISSUE_FLAGS_FORCE_BACKTRACE), "GST_VALIDATE_ISSUE_FLAGS_FORCE_BACKTRACE", "force-backtrace" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_flags_register_static ("GstValidateIssueFlags", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "gst-validate-reporter.h" */
GType
gst_validate_interception_return_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(GST_VALIDATE_REPORTER_DROP), "GST_VALIDATE_REPORTER_DROP", "drop" },
    { C_ENUM(GST_VALIDATE_REPORTER_KEEP), "GST_VALIDATE_REPORTER_KEEP", "keep" },
    { C_ENUM(GST_VALIDATE_REPORTER_REPORT), "GST_VALIDATE_REPORTER_REPORT", "report" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("GstValidateInterceptionReturn", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "gst-validate-scenario.h" */
GType
gst_validate_action_return_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(GST_VALIDATE_EXECUTE_ACTION_ERROR), "GST_VALIDATE_EXECUTE_ACTION_ERROR", "error" },
    { C_ENUM(GST_VALIDATE_EXECUTE_ACTION_OK), "GST_VALIDATE_EXECUTE_ACTION_OK", "ok" },
    { C_ENUM(GST_VALIDATE_EXECUTE_ACTION_ASYNC), "GST_VALIDATE_EXECUTE_ACTION_ASYNC", "async" },
    { C_ENUM(GST_VALIDATE_EXECUTE_ACTION_NON_BLOCKING), "GST_VALIDATE_EXECUTE_ACTION_NON_BLOCKING", "non-blocking" },
    { C_ENUM(GST_VALIDATE_EXECUTE_ACTION_INTERLACED), "GST_VALIDATE_EXECUTE_ACTION_INTERLACED", "interlaced" },
    { C_ENUM(GST_VALIDATE_EXECUTE_ACTION_ERROR_REPORTED), "GST_VALIDATE_EXECUTE_ACTION_ERROR_REPORTED", "error-reported" },
    { C_ENUM(GST_VALIDATE_EXECUTE_ACTION_IN_PROGRESS), "GST_VALIDATE_EXECUTE_ACTION_IN_PROGRESS", "in-progress" },
    { C_ENUM(GST_VALIDATE_EXECUTE_ACTION_NONE), "GST_VALIDATE_EXECUTE_ACTION_NONE", "none" },
    { C_ENUM(GST_VALIDATE_EXECUTE_ACTION_DONE), "GST_VALIDATE_EXECUTE_ACTION_DONE", "done" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("GstValidateActionReturn", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
gst_validate_action_type_flags_get_type (void)
{
  static gsize id = 0;
  static const GFlagsValue values[] = {
    { C_FLAGS(GST_VALIDATE_ACTION_TYPE_NONE), "GST_VALIDATE_ACTION_TYPE_NONE", "none" },
    { C_FLAGS(GST_VALIDATE_ACTION_TYPE_CONFIG), "GST_VALIDATE_ACTION_TYPE_CONFIG", "config" },
    { C_FLAGS(GST_VALIDATE_ACTION_TYPE_ASYNC), "GST_VALIDATE_ACTION_TYPE_ASYNC", "async" },
    { C_FLAGS(GST_VALIDATE_ACTION_TYPE_NON_BLOCKING), "GST_VALIDATE_ACTION_TYPE_NON_BLOCKING", "non-blocking" },
    { C_FLAGS(GST_VALIDATE_ACTION_TYPE_INTERLACED), "GST_VALIDATE_ACTION_TYPE_INTERLACED", "interlaced" },
    { C_FLAGS(GST_VALIDATE_ACTION_TYPE_CAN_EXECUTE_ON_ADDITION), "GST_VALIDATE_ACTION_TYPE_CAN_EXECUTE_ON_ADDITION", "can-execute-on-addition" },
    { C_FLAGS(GST_VALIDATE_ACTION_TYPE_NEEDS_CLOCK), "GST_VALIDATE_ACTION_TYPE_NEEDS_CLOCK", "needs-clock" },
    { C_FLAGS(GST_VALIDATE_ACTION_TYPE_NO_EXECUTION_NOT_FATAL), "GST_VALIDATE_ACTION_TYPE_NO_EXECUTION_NOT_FATAL", "no-execution-not-fatal" },
    { C_FLAGS(GST_VALIDATE_ACTION_TYPE_CAN_BE_OPTIONAL), "GST_VALIDATE_ACTION_TYPE_CAN_BE_OPTIONAL", "can-be-optional" },
    { C_FLAGS(GST_VALIDATE_ACTION_TYPE_DOESNT_NEED_PIPELINE), "GST_VALIDATE_ACTION_TYPE_DOESNT_NEED_PIPELINE", "doesnt-need-pipeline" },
    { C_FLAGS(GST_VALIDATE_ACTION_TYPE_HANDLED_IN_CONFIG), "GST_VALIDATE_ACTION_TYPE_HANDLED_IN_CONFIG", "handled-in-config" },
    { C_FLAGS(GST_VALIDATE_ACTION_TYPE_CHECK), "GST_VALIDATE_ACTION_TYPE_CHECK", "check" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_flags_register_static ("GstValidateActionTypeFlags", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "media-descriptor-writer.h" */
GType
gst_validate_media_descriptor_writer_flags_get_type (void)
{
  static gsize id = 0;
  static const GFlagsValue values[] = {
    { C_FLAGS(GST_VALIDATE_MEDIA_DESCRIPTOR_WRITER_FLAGS_NONE), "GST_VALIDATE_MEDIA_DESCRIPTOR_WRITER_FLAGS_NONE", "none" },
    { C_FLAGS(GST_VALIDATE_MEDIA_DESCRIPTOR_WRITER_FLAGS_NO_PARSER), "GST_VALIDATE_MEDIA_DESCRIPTOR_WRITER_FLAGS_NO_PARSER", "no-parser" },
    { C_FLAGS(GST_VALIDATE_MEDIA_DESCRIPTOR_WRITER_FLAGS_FULL), "GST_VALIDATE_MEDIA_DESCRIPTOR_WRITER_FLAGS_FULL", "full" },
    { C_FLAGS(GST_VALIDATE_MEDIA_DESCRIPTOR_WRITER_FLAGS_HANDLE_GLOGS), "GST_VALIDATE_MEDIA_DESCRIPTOR_WRITER_FLAGS_HANDLE_GLOGS", "handle-glogs" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_flags_register_static ("GstValidateMediaDescriptorWriterFlags", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}


/* Generated data ends here */

