
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "gl-enumtypes.h"

#include <gst/gl/gl.h>

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gstgl_enums.h" */
GType
gst_gl_texture_target_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_GL_TEXTURE_TARGET_NONE), "GST_GL_TEXTURE_TARGET_NONE", "none" },
      { C_ENUM (GST_GL_TEXTURE_TARGET_2D), "GST_GL_TEXTURE_TARGET_2D", "2d" },
      { C_ENUM (GST_GL_TEXTURE_TARGET_RECTANGLE), "GST_GL_TEXTURE_TARGET_RECTANGLE", "rectangle" },
      { C_ENUM (GST_GL_TEXTURE_TARGET_EXTERNAL_OES), "GST_GL_TEXTURE_TARGET_EXTERNAL_OES", "external-oes" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstGLTextureTarget", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstglapi.h" */
GType
gst_gl_api_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GFlagsValue values[] = {
      { C_FLAGS (GST_GL_API_NONE), "GST_GL_API_NONE", "none" },
      { C_FLAGS (GST_GL_API_OPENGL), "GST_GL_API_OPENGL", "opengl" },
      { C_FLAGS (GST_GL_API_OPENGL3), "GST_GL_API_OPENGL3", "opengl3" },
      { C_FLAGS (GST_GL_API_GLES1), "GST_GL_API_GLES1", "gles1" },
      { C_FLAGS (GST_GL_API_GLES2), "GST_GL_API_GLES2", "gles2" },
      { C_FLAGS (GST_GL_API_ANY), "GST_GL_API_ANY", "any" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_flags_register_static ("GstGLAPI", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}
GType
gst_gl_platform_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GFlagsValue values[] = {
      { C_FLAGS (GST_GL_PLATFORM_NONE), "GST_GL_PLATFORM_NONE", "none" },
      { C_FLAGS (GST_GL_PLATFORM_EGL), "GST_GL_PLATFORM_EGL", "egl" },
      { C_FLAGS (GST_GL_PLATFORM_GLX), "GST_GL_PLATFORM_GLX", "glx" },
      { C_FLAGS (GST_GL_PLATFORM_WGL), "GST_GL_PLATFORM_WGL", "wgl" },
      { C_FLAGS (GST_GL_PLATFORM_CGL), "GST_GL_PLATFORM_CGL", "cgl" },
      { C_FLAGS (GST_GL_PLATFORM_EAGL), "GST_GL_PLATFORM_EAGL", "eagl" },
      { C_FLAGS (GST_GL_PLATFORM_ANY), "GST_GL_PLATFORM_ANY", "any" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_flags_register_static ("GstGLPlatform", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstglbasememory.h" */
GType
gst_gl_base_memory_error_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_GL_BASE_MEMORY_ERROR_FAILED), "GST_GL_BASE_MEMORY_ERROR_FAILED", "failed" },
      { C_ENUM (GST_GL_BASE_MEMORY_ERROR_OLD_LIBS), "GST_GL_BASE_MEMORY_ERROR_OLD_LIBS", "old-libs" },
      { C_ENUM (GST_GL_BASE_MEMORY_ERROR_RESOURCE_UNAVAILABLE), "GST_GL_BASE_MEMORY_ERROR_RESOURCE_UNAVAILABLE", "resource-unavailable" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstGLBaseMemoryError", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}
GType
gst_gl_base_memory_transfer_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GFlagsValue values[] = {
      { C_FLAGS (GST_GL_BASE_MEMORY_TRANSFER_NEED_DOWNLOAD), "GST_GL_BASE_MEMORY_TRANSFER_NEED_DOWNLOAD", "download" },
      { C_FLAGS (GST_GL_BASE_MEMORY_TRANSFER_NEED_UPLOAD), "GST_GL_BASE_MEMORY_TRANSFER_NEED_UPLOAD", "upload" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_flags_register_static ("GstGLBaseMemoryTransfer", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstglcontext.h" */
GType
gst_gl_context_error_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_GL_CONTEXT_ERROR_FAILED), "GST_GL_CONTEXT_ERROR_FAILED", "failed" },
      { C_ENUM (GST_GL_CONTEXT_ERROR_WRONG_CONFIG), "GST_GL_CONTEXT_ERROR_WRONG_CONFIG", "wrong-config" },
      { C_ENUM (GST_GL_CONTEXT_ERROR_WRONG_API), "GST_GL_CONTEXT_ERROR_WRONG_API", "wrong-api" },
      { C_ENUM (GST_GL_CONTEXT_ERROR_OLD_LIBS), "GST_GL_CONTEXT_ERROR_OLD_LIBS", "old-libs" },
      { C_ENUM (GST_GL_CONTEXT_ERROR_CREATE_CONTEXT), "GST_GL_CONTEXT_ERROR_CREATE_CONTEXT", "create-context" },
      { C_ENUM (GST_GL_CONTEXT_ERROR_RESOURCE_UNAVAILABLE), "GST_GL_CONTEXT_ERROR_RESOURCE_UNAVAILABLE", "resource-unavailable" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstGLContextError", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstglcontextconfig.h" */
GType
gst_gl_config_caveat_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_GL_CONFIG_CAVEAT_NONE), "GST_GL_CONFIG_CAVEAT_NONE", "none" },
      { C_ENUM (GST_GL_CONFIG_CAVEAT_SLOW), "GST_GL_CONFIG_CAVEAT_SLOW", "slow" },
      { C_ENUM (GST_GL_CONFIG_CAVEAT_NON_CONFORMANT), "GST_GL_CONFIG_CAVEAT_NON_CONFORMANT", "non-conformant" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstGLConfigCaveat", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}
GType
gst_gl_config_surface_type_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GFlagsValue values[] = {
      { C_FLAGS (GST_GL_CONFIG_SURFACE_TYPE_NONE), "GST_GL_CONFIG_SURFACE_TYPE_NONE", "none" },
      { C_FLAGS (GST_GL_CONFIG_SURFACE_TYPE_WINDOW), "GST_GL_CONFIG_SURFACE_TYPE_WINDOW", "window" },
      { C_FLAGS (GST_GL_CONFIG_SURFACE_TYPE_PBUFFER), "GST_GL_CONFIG_SURFACE_TYPE_PBUFFER", "pbuffer" },
      { C_FLAGS (GST_GL_CONFIG_SURFACE_TYPE_PIXMAP), "GST_GL_CONFIG_SURFACE_TYPE_PIXMAP", "pixmap" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_flags_register_static ("GstGLConfigSurfaceType", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstgldisplay.h" */
GType
gst_gl_display_type_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GFlagsValue values[] = {
      { C_FLAGS (GST_GL_DISPLAY_TYPE_NONE), "GST_GL_DISPLAY_TYPE_NONE", "none" },
      { C_FLAGS (GST_GL_DISPLAY_TYPE_X11), "GST_GL_DISPLAY_TYPE_X11", "x11" },
      { C_FLAGS (GST_GL_DISPLAY_TYPE_WAYLAND), "GST_GL_DISPLAY_TYPE_WAYLAND", "wayland" },
      { C_FLAGS (GST_GL_DISPLAY_TYPE_COCOA), "GST_GL_DISPLAY_TYPE_COCOA", "cocoa" },
      { C_FLAGS (GST_GL_DISPLAY_TYPE_WIN32), "GST_GL_DISPLAY_TYPE_WIN32", "win32" },
      { C_FLAGS (GST_GL_DISPLAY_TYPE_DISPMANX), "GST_GL_DISPLAY_TYPE_DISPMANX", "dispmanx" },
      { C_FLAGS (GST_GL_DISPLAY_TYPE_EGL), "GST_GL_DISPLAY_TYPE_EGL", "egl" },
      { C_FLAGS (GST_GL_DISPLAY_TYPE_VIV_FB), "GST_GL_DISPLAY_TYPE_VIV_FB", "viv-fb" },
      { C_FLAGS (GST_GL_DISPLAY_TYPE_GBM), "GST_GL_DISPLAY_TYPE_GBM", "gbm" },
      { C_FLAGS (GST_GL_DISPLAY_TYPE_EGL_DEVICE), "GST_GL_DISPLAY_TYPE_EGL_DEVICE", "egl-device" },
      { C_FLAGS (GST_GL_DISPLAY_TYPE_EAGL), "GST_GL_DISPLAY_TYPE_EAGL", "eagl" },
      { C_FLAGS (GST_GL_DISPLAY_TYPE_WINRT), "GST_GL_DISPLAY_TYPE_WINRT", "winrt" },
      { C_FLAGS (GST_GL_DISPLAY_TYPE_ANDROID), "GST_GL_DISPLAY_TYPE_ANDROID", "android" },
      { C_FLAGS (GST_GL_DISPLAY_TYPE_EGL_SURFACELESS), "GST_GL_DISPLAY_TYPE_EGL_SURFACELESS", "egl-surfaceless" },
      { C_FLAGS (GST_GL_DISPLAY_TYPE_ANY), "GST_GL_DISPLAY_TYPE_ANY", "any" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_flags_register_static ("GstGLDisplayType", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstglformat.h" */
GType
gst_gl_format_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_GL_LUMINANCE), "GST_GL_LUMINANCE", "luminance" },
      { C_ENUM (GST_GL_ALPHA), "GST_GL_ALPHA", "alpha" },
      { C_ENUM (GST_GL_LUMINANCE_ALPHA), "GST_GL_LUMINANCE_ALPHA", "luminance-alpha" },
      { C_ENUM (GST_GL_RED), "GST_GL_RED", "red" },
      { C_ENUM (GST_GL_R8), "GST_GL_R8", "r8" },
      { C_ENUM (GST_GL_RG), "GST_GL_RG", "rg" },
      { C_ENUM (GST_GL_RG8), "GST_GL_RG8", "rg8" },
      { C_ENUM (GST_GL_RGB), "GST_GL_RGB", "rgb" },
      { C_ENUM (GST_GL_RGB8), "GST_GL_RGB8", "rgb8" },
      { C_ENUM (GST_GL_RGB565), "GST_GL_RGB565", "rgb565" },
      { C_ENUM (GST_GL_RGB16), "GST_GL_RGB16", "rgb16" },
      { C_ENUM (GST_GL_RGBA), "GST_GL_RGBA", "rgba" },
      { C_ENUM (GST_GL_RGBA8), "GST_GL_RGBA8", "rgba8" },
      { C_ENUM (GST_GL_RGBA16), "GST_GL_RGBA16", "rgba16" },
      { C_ENUM (GST_GL_DEPTH_COMPONENT16), "GST_GL_DEPTH_COMPONENT16", "depth-component16" },
      { C_ENUM (GST_GL_DEPTH24_STENCIL8), "GST_GL_DEPTH24_STENCIL8", "depth24-stencil8" },
      { C_ENUM (GST_GL_RGB10_A2), "GST_GL_RGB10_A2", "rgb10-a2" },
      { C_ENUM (GST_GL_R16), "GST_GL_R16", "r16" },
      { C_ENUM (GST_GL_RG16), "GST_GL_RG16", "rg16" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstGLFormat", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstglquery.h" */
GType
gst_gl_query_type_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_GL_QUERY_NONE), "GST_GL_QUERY_NONE", "none" },
      { C_ENUM (GST_GL_QUERY_TIME_ELAPSED), "GST_GL_QUERY_TIME_ELAPSED", "time-elapsed" },
      { C_ENUM (GST_GL_QUERY_TIMESTAMP), "GST_GL_QUERY_TIMESTAMP", "timestamp" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstGLQueryType", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstglsl.h" */
GType
gst_glsl_error_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_GLSL_ERROR_COMPILE), "GST_GLSL_ERROR_COMPILE", "compile" },
      { C_ENUM (GST_GLSL_ERROR_LINK), "GST_GLSL_ERROR_LINK", "link" },
      { C_ENUM (GST_GLSL_ERROR_PROGRAM), "GST_GLSL_ERROR_PROGRAM", "program" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstGLSLError", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}
GType
gst_glsl_version_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_GLSL_VERSION_NONE), "GST_GLSL_VERSION_NONE", "none" },
      { C_ENUM (GST_GLSL_VERSION_100), "GST_GLSL_VERSION_100", "100" },
      { C_ENUM (GST_GLSL_VERSION_110), "GST_GLSL_VERSION_110", "110" },
      { C_ENUM (GST_GLSL_VERSION_120), "GST_GLSL_VERSION_120", "120" },
      { C_ENUM (GST_GLSL_VERSION_130), "GST_GLSL_VERSION_130", "130" },
      { C_ENUM (GST_GLSL_VERSION_140), "GST_GLSL_VERSION_140", "140" },
      { C_ENUM (GST_GLSL_VERSION_150), "GST_GLSL_VERSION_150", "150" },
      { C_ENUM (GST_GLSL_VERSION_300), "GST_GLSL_VERSION_300", "300" },
      { C_ENUM (GST_GLSL_VERSION_310), "GST_GLSL_VERSION_310", "310" },
      { C_ENUM (GST_GLSL_VERSION_320), "GST_GLSL_VERSION_320", "320" },
      { C_ENUM (GST_GLSL_VERSION_330), "GST_GLSL_VERSION_330", "330" },
      { C_ENUM (GST_GLSL_VERSION_400), "GST_GLSL_VERSION_400", "400" },
      { C_ENUM (GST_GLSL_VERSION_410), "GST_GLSL_VERSION_410", "410" },
      { C_ENUM (GST_GLSL_VERSION_420), "GST_GLSL_VERSION_420", "420" },
      { C_ENUM (GST_GLSL_VERSION_430), "GST_GLSL_VERSION_430", "430" },
      { C_ENUM (GST_GLSL_VERSION_440), "GST_GLSL_VERSION_440", "440" },
      { C_ENUM (GST_GLSL_VERSION_450), "GST_GLSL_VERSION_450", "450" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstGLSLVersion", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}
GType
gst_glsl_profile_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GFlagsValue values[] = {
      { C_FLAGS (GST_GLSL_PROFILE_NONE), "GST_GLSL_PROFILE_NONE", "none" },
      { C_FLAGS (GST_GLSL_PROFILE_ES), "GST_GLSL_PROFILE_ES", "es" },
      { C_FLAGS (GST_GLSL_PROFILE_CORE), "GST_GLSL_PROFILE_CORE", "core" },
      { C_FLAGS (GST_GLSL_PROFILE_COMPATIBILITY), "GST_GLSL_PROFILE_COMPATIBILITY", "compatibility" },
      { C_FLAGS (GST_GLSL_PROFILE_ANY), "GST_GLSL_PROFILE_ANY", "any" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_flags_register_static ("GstGLSLProfile", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstglupload.h" */
GType
gst_gl_upload_return_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_GL_UPLOAD_DONE), "GST_GL_UPLOAD_DONE", "done" },
      { C_ENUM (GST_GL_UPLOAD_ERROR), "GST_GL_UPLOAD_ERROR", "error" },
      { C_ENUM (GST_GL_UPLOAD_UNSUPPORTED), "GST_GL_UPLOAD_UNSUPPORTED", "unsupported" },
      { C_ENUM (GST_GL_UPLOAD_RECONFIGURE), "GST_GL_UPLOAD_RECONFIGURE", "reconfigure" },
      { C_ENUM (GST_GL_UPLOAD_UNSHARED_GL_CONTEXT), "GST_GL_UPLOAD_UNSHARED_GL_CONTEXT", "unshared-gl-context" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstGLUploadReturn", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstglutils.h" */
GType
gst_gl_drm_format_flags_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GFlagsValue values[] = {
      { C_FLAGS (GST_GL_DRM_FORMAT_INCLUDE_EXTERNAL), "GST_GL_DRM_FORMAT_INCLUDE_EXTERNAL", "include-external" },
      { C_FLAGS (GST_GL_DRM_FORMAT_LINEAR_ONLY), "GST_GL_DRM_FORMAT_LINEAR_ONLY", "linear-only" },
      { C_FLAGS (GST_GL_DRM_FORMAT_INCLUDE_EMULATED), "GST_GL_DRM_FORMAT_INCLUDE_EMULATED", "include-emulated" },
      { C_FLAGS (GST_GL_DRM_FORMAT_DIRECT_IMPORT), "GST_GL_DRM_FORMAT_DIRECT_IMPORT", "direct-import" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_flags_register_static ("GstGLDrmFormatFlags", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstglviewconvert.h" */
GType
gst_gl_stereo_downmix_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_GL_STEREO_DOWNMIX_ANAGLYPH_GREEN_MAGENTA_DUBOIS), "GST_GL_STEREO_DOWNMIX_ANAGLYPH_GREEN_MAGENTA_DUBOIS", "green-magenta-dubois" },
      { C_ENUM (GST_GL_STEREO_DOWNMIX_ANAGLYPH_RED_CYAN_DUBOIS), "GST_GL_STEREO_DOWNMIX_ANAGLYPH_RED_CYAN_DUBOIS", "red-cyan-dubois" },
      { C_ENUM (GST_GL_STEREO_DOWNMIX_ANAGLYPH_AMBER_BLUE_DUBOIS), "GST_GL_STEREO_DOWNMIX_ANAGLYPH_AMBER_BLUE_DUBOIS", "amber-blue-dubois" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstGLStereoDownmix", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* enumerations from "gstglwindow.h" */
GType
gst_gl_window_error_get_type (void)
{
  static gsize static_g_define_type_id = 0;
  if (g_once_init_enter (&static_g_define_type_id)) {
    static const GEnumValue values[] = {
      { C_ENUM (GST_GL_WINDOW_ERROR_FAILED), "GST_GL_WINDOW_ERROR_FAILED", "failed" },
      { C_ENUM (GST_GL_WINDOW_ERROR_OLD_LIBS), "GST_GL_WINDOW_ERROR_OLD_LIBS", "old-libs" },
      { C_ENUM (GST_GL_WINDOW_ERROR_RESOURCE_UNAVAILABLE), "GST_GL_WINDOW_ERROR_RESOURCE_UNAVAILABLE", "resource-unavailable" },
      { 0, NULL, NULL }
    };
    GType g_define_type_id = g_enum_register_static ("GstGLWindowError", values);
    g_once_init_leave (&static_g_define_type_id, g_define_type_id);
  }
  return static_g_define_type_id;
}

/* Generated data ends here */

