
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "gstvpx-enumtypes.h"
#include "gstvpxenums.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gstvpxenums.h" */

GType
gst_vpxaq_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GST_VPX_AQ_OFF), "GST_VPX_AQ_OFF", "off" },
        { C_ENUM (GST_VPX_AQ_VARIANCE), "GST_VPX_AQ_VARIANCE", "variance" },
        { C_ENUM (GST_VPX_AQ_COMPLEXITY), "GST_VPX_AQ_COMPLEXITY", "complexity" },
        { C_ENUM (GST_VPX_AQ_CYCLIC_REFRESH), "GST_VPX_AQ_CYCLIC_REFRESH", "cyclic-refresh" },
        { C_ENUM (GST_VPX_AQ_EQUATOR360), "GST_VPX_AQ_EQUATOR360", "equator360" },
        { C_ENUM (GST_VPX_AQ_PERCEPTUAL), "GST_VPX_AQ_PERCEPTUAL", "perceptual" },
        { C_ENUM (GST_VPX_AQ_PSNR), "GST_VPX_AQ_PSNR", "psnr" },
        { C_ENUM (GST_VPX_AQ_LOOKAHEAD), "GST_VPX_AQ_LOOKAHEAD", "lookahead" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GstVPXAQ"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

