
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "mse-enumtypes.h"
#include "gstmediasource.h"
#include "gstsourcebuffer.h"
#include "gstsourcebufferlist.h"
#include "gstmsesrc.h"
#include "mse.h"
#include "mse-prelude.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gstmediasource.h" */

GType
gst_media_source_ready_state_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GST_MEDIA_SOURCE_READY_STATE_CLOSED), "GST_MEDIA_SOURCE_READY_STATE_CLOSED", "closed" },
        { C_ENUM (GST_MEDIA_SOURCE_READY_STATE_OPEN), "GST_MEDIA_SOURCE_READY_STATE_OPEN", "open" },
        { C_ENUM (GST_MEDIA_SOURCE_READY_STATE_ENDED), "GST_MEDIA_SOURCE_READY_STATE_ENDED", "ended" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GstMediaSourceReadyState"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gst_media_source_error_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GST_MEDIA_SOURCE_ERROR_INVALID_STATE), "GST_MEDIA_SOURCE_ERROR_INVALID_STATE", "invalid-state" },
        { C_ENUM (GST_MEDIA_SOURCE_ERROR_TYPE), "GST_MEDIA_SOURCE_ERROR_TYPE", "type" },
        { C_ENUM (GST_MEDIA_SOURCE_ERROR_NOT_SUPPORTED), "GST_MEDIA_SOURCE_ERROR_NOT_SUPPORTED", "not-supported" },
        { C_ENUM (GST_MEDIA_SOURCE_ERROR_NOT_FOUND), "GST_MEDIA_SOURCE_ERROR_NOT_FOUND", "not-found" },
        { C_ENUM (GST_MEDIA_SOURCE_ERROR_QUOTA_EXCEEDED), "GST_MEDIA_SOURCE_ERROR_QUOTA_EXCEEDED", "quota-exceeded" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GstMediaSourceError"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gst_media_source_eos_error_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GST_MEDIA_SOURCE_EOS_ERROR_NONE), "GST_MEDIA_SOURCE_EOS_ERROR_NONE", "none" },
        { C_ENUM (GST_MEDIA_SOURCE_EOS_ERROR_NETWORK), "GST_MEDIA_SOURCE_EOS_ERROR_NETWORK", "network" },
        { C_ENUM (GST_MEDIA_SOURCE_EOS_ERROR_DECODE), "GST_MEDIA_SOURCE_EOS_ERROR_DECODE", "decode" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GstMediaSourceEOSError"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "gstmsesrc.h" */

GType
gst_mse_src_ready_state_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GST_MSE_SRC_READY_STATE_HAVE_NOTHING), "GST_MSE_SRC_READY_STATE_HAVE_NOTHING", "nothing" },
        { C_ENUM (GST_MSE_SRC_READY_STATE_HAVE_METADATA), "GST_MSE_SRC_READY_STATE_HAVE_METADATA", "metadata" },
        { C_ENUM (GST_MSE_SRC_READY_STATE_HAVE_CURRENT_DATA), "GST_MSE_SRC_READY_STATE_HAVE_CURRENT_DATA", "current-data" },
        { C_ENUM (GST_MSE_SRC_READY_STATE_HAVE_FUTURE_DATA), "GST_MSE_SRC_READY_STATE_HAVE_FUTURE_DATA", "future-data" },
        { C_ENUM (GST_MSE_SRC_READY_STATE_HAVE_ENOUGH_DATA), "GST_MSE_SRC_READY_STATE_HAVE_ENOUGH_DATA", "enough-data" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GstMseSrcReadyState"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "gstsourcebuffer.h" */

GType
gst_source_buffer_append_mode_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GST_SOURCE_BUFFER_APPEND_MODE_SEGMENTS), "GST_SOURCE_BUFFER_APPEND_MODE_SEGMENTS", "segments" },
        { C_ENUM (GST_SOURCE_BUFFER_APPEND_MODE_SEQUENCE), "GST_SOURCE_BUFFER_APPEND_MODE_SEQUENCE", "sequence" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GstSourceBufferAppendMode"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

